import asyncio
from hashlib import sha1
import random
import os

import aiofiles
from tqdm import tqdm

import bencode
from tracker import Tracker
from peer_manager import PeerManager

from string import digits, ascii_letters
chars = digits + ascii_letters

class Torrent:
    def __init__(self, torrent_path, max_peers=15):
        self.peer_id = ''.join(random.choice(chars) for i in range(20))
        self.max_peers = max_peers
        with open(torrent_path, 'rb') as f:
            self.torrent_data = bencode.decode(f.read())
        self.info_hash = sha1(bencode.encode(self.torrent_data['info'])).digest()
        self.piece_length = self.torrent_data['info']['piece length']
        self.total_length = 0
        self.files = []
        if 'files' in self.torrent_data['info']:
            folder = self.torrent_data['info'].get('name', torrent_path[:torrent_path.rfind('.')])
            os.makedirs(folder, exist_ok=True)
            for file in self.torrent_data['info']['files']:
                for i in range(len(file['path']) - 1):
                    os.makedirs(os.path.join(folder, *file['path'][:i+1]), exist_ok=True)
                if not os.path.exists(os.path.join(folder, *file['path'])):
                    open(os.path.join(folder, *file['path']), 'w').close()
                self.files.append({'path': os.path.join(folder, *file['path']), 'length': file['length']})
                self.total_length += file['length']
        else:
            name = self.torrent_data['info'].get('name', torrent_path[:torrent_path.rfind('.')])
            open(name, 'w').close()
            self.files.append({'path': name, 'length': self.torrent_data['info']['length']})
            self.total_length = self.torrent_data['info']['length']
        self.piece_count = (self.total_length + self.piece_length - 1) // self.piece_length
        trackers = [Tracker(addr[0], self) for addr in
                         [[self.torrent_data['announce']]] + self.torrent_data.get('announce-list', [])]
        self.peer_manager = PeerManager(trackers, self, max_peers)
        self.pieces = {i: [(0, self.piece_length)]
                       for i in range(self.piece_count)}
        self.pieces[self.piece_count - 1] = [(0, self.total_length % self.piece_length)]
        self.in_progress = {}
        self.done = []
    
    async def _check_pieces(self):
        cur_piece = 0
        data = b''
        for file in self.files:
            async with aiofiles.open(file['path'], 'r+b') as f:
                await f.truncate(file['length'])
                while True:
                    data += await f.read(self.piece_length - len(data))
                    if len(data) != self.piece_length:
                        break
                    if sha1(data).digest() == self.torrent_data['info']['pieces'][cur_piece * 20: (cur_piece+1) * 20]:
                        # print('\033[96m' + 'already got piece', cur_piece, '\033[0m')
                        del self.pieces[cur_piece]
                        self.done.append(cur_piece)
                        self.progress_bar.update(self.piece_length)
                        self._update_progress_bar()
                    data = b''
                    cur_piece += 1
        if sha1(data).digest() == self.torrent_data['info']['pieces'][cur_piece * 20: (cur_piece+1) * 20]:
            # print('\033[96m' + 'already got piece', cur_piece, '\033[0m')
            del self.pieces[cur_piece]
            self.done.append(cur_piece)
            self.progress_bar.update(len(data))
            self._update_progress_bar()
    
    def _update_progress_bar(self):
        self.progress_bar.set_description(f'Peers {len(self.peer_manager.active_peers)} / {self.max_peers}'
                                               f' | Pieces {len(self.done)} / {self.piece_count} ')
    
    async def start(self):
        self.progress_bar = tqdm(total=self.total_length, 
                                 unit='B', unit_scale=True, 
                                 desc=f'Peers 0 / {self.max_peers} Pieces 0 / {self.piece_count} ',
                                 ncols=90)
        await self._check_pieces()
        # print('REMAINING', len(self.pieces), 'from', self.piece_count)
        asyncio.create_task(self.peer_manager.ensure_peers())
        while self.pieces:
            for i in self.pieces:
                if i in self.in_progress and len(self.pieces) != len(self.in_progress):
                    continue
                await self.peer_manager.wait_ready()
                peer = await self.peer_manager.peer_having_piece(i)
                if peer:
                    # print('\033[93m' + 'got peer', i, self.peer_manager.active_peers.__len__(), '\033[0m')
                    piece_corutine = await self.peer_manager.get_piece_from_peer(i, peer, self.pieces[i][-1][1])
                    self.in_progress[i] = asyncio.create_task(self._new_piece(i, piece_corutine))
                    self._update_progress_bar()
            for i in self.done:
                self.pieces.pop(i, None)
                self.in_progress.pop(i, None)
            await self.peer_manager.remove_peers_not_having_piece(self.pieces.keys())
    
    async def _new_piece(self, index, piece_corut):
        try:
            data = await asyncio.wait_for(piece_corut, 60) #TODO: pieces timeout
            # print('\033[92m' + 'new', index)
            if sha1(data).digest() == self.torrent_data['info']['pieces'][index * 20: (index+1) * 20]:
                # print('\033[92m' + 'new', index, '\033[0m')
                start = self.piece_length * index
                end = start + len(data)
                cur = 0
                for file in self.files:
                    cur += file['length']
                    if start <= cur:
                        async with aiofiles.open(file['path'], 'r+b') as f:
                            await f.seek(start - cur + file['length'])
                            if end > cur:
                                await f.write(data[: cur - start])
                                data = data[cur - start:]
                                start = cur
                            else:
                                await f.write(data)
                                break
                self.done.append(index)
                self.progress_bar.update(len(data))
                self._update_progress_bar()
            else:
                self.in_progress.pop(index, None)
                # self.progress_bar.update(len(data))
                # print('\033[91m' + 'bad hash', index, '\033[0m')
        except asyncio.TimeoutError:
            self.in_progress.pop(index, None)



import asyncio
import struct

HANDSHAKE_TIMEOUT = 15
KEEP_ALIVE_TIMEOUT = 60

class Peer():
    def __init__(self, ip, port, torrent):
        self.ip = ip
        self.port = port
        self.torrent = torrent
        self.am_choking = True
        self.am_interested = False
        self.peer_choking = True
        self.peer_interested = False
        self.healthy = False
        self.pieces = [False] * torrent.piece_count
        self.peer_id = None
        self.writer = None
        self.reader = None
        self.last_active = 0
        self.keep_alive_task = None
        self.events = {}
    
    async def handshake(self):
        pstrlen = 19
        pstr = b"BitTorrent protocol"
        reserved = 0
        info_hash = self.torrent.info_hash
        peer_id = self.torrent.peer_id
        message = struct.pack(f'>B{pstrlen}sQ', pstrlen, pstr, reserved) + info_hash + bytes(peer_id, 'utf8')

        async def handle_handshake():
            reader, writer = await asyncio.open_connection(self.ip, self.port)
            writer.write(message)
            await writer.drain()
            # print('wrote')
            data = await reader.readexactly(1)
            # print(f'Received: {data.decode()!r}')
            data = int.from_bytes(data, 'big')
            newdata = await reader.readexactly(data)
            # print('pstr', newdata.decode())
            newdata = await reader.readexactly(8)
            newdata = await reader.readexactly(20)
            peer_id = await reader.readexactly(20)
            self.peer_id = peer_id
            self.writer = writer
            self.reader = reader

            return info_hash
        if self.torrent.info_hash != (await asyncio.wait_for(handle_handshake(), timeout=HANDSHAKE_TIMEOUT)):
            return False

        self.healthy = True
        self.last_active = asyncio.get_event_loop().time()
        #TODO: gives error
        # self.keep_alive_task = asyncio.get_event_loop().call_later(KEEP_ALIVE_TIMEOUT, asyncio.create_task, self.keep_alive())
        # print('after schedule')
        asyncio.create_task(self._listen())
        return True
        
        # print('Close the connection')
        # writer.close()
        # await writer.wait_closed()
    
    async def _listen(self):
        # print('entered listen mode')
        while self.healthy:
            try:
                data = await self.reader.readexactly(4)
                length = int.from_bytes(data, 'big')
                if length == 0: # Keep alive
                    continue

                data = await self.reader.readexactly(1)
                message_id = int.from_bytes(data, 'big')

                length -= 1

                payload = await self.reader.readexactly(length)
                await self._handle_message(message_id, payload)
            except Exception as e:
                # print('exception in listen', e, type(e))
                await self.drop()
                break
    
    async def _handle_message(self, message_id, payload):
        # print('message_id', message_id)
        match message_id:
            case 0: # Choke
                self.am_choking = True
            case 1: # Unchoke
                # print('unchoked')
                if 'unchoked' in self.events:
                    self.events['unchoked'].set()
                    del self.events['unchoked']
                self.am_choking = False
            case 2: # Interested
                self.peer_interested = True
            case 3: # NotInterested
                self.peer_interested = False
            case 4: # Have
                has_index = int.from_bytes(payload, 'big')
                self.pieces[has_index] = True
            case 5: # Bitfield
                if len(payload) != ((self.torrent.piece_count + 8 - 1) // 8):
                    await self.drop()
                    return
                
                index = 0
                for byte in payload:
                    have = bin(byte)[2:]
                    for bit in have:
                        if bit == '1':
                            if index < len(self.pieces):
                                self.pieces[index] = True
                            else:
                                # some error happend
                                await self.drop()
                        index += 1
                # print('received bitfield')
            case 6: # Request
                index, begin, length = struct.unpack('>III', payload)
            case 7: # Piece
                index, begin = struct.unpack('>II', payload[:8])
                block = payload[8:]
                # print(index, begin, self.events)
                if (index, begin) in self.events:
                    self.events[index, begin].set_result(block)
                    del self.events[index, begin]
                # await self.torrent.new_block(index, begin, block)
            case 8: # Cancel
                index, begin, length = struct.unpack('>III', payload)
            case 9: # Port
                port = int.from_bytes(payload, 'big')
    
    async def request_block(self, index, begin, length, event:asyncio.Future=None):
        try:
            self.last_active = asyncio.get_event_loop().time()
            # self.keep_alive_task.cancel()
            if event:
                self.events[index, begin] = event
            data = struct.pack('>IBIII', 13, 6, index, begin, length)  # <len=13><id=6>
            self.writer.write(data)
            await self.writer.drain()
            # self.keep_alive_task = asyncio.get_event_loop().call_later(KEEP_ALIVE_TIMEOUT, asyncio.create_task, self.keep_alive())
        except Exception as e:
            await self.drop()

    async def show_interest(self, event:asyncio.Event=None):
        try:
            self.last_active = asyncio.get_event_loop().time()
            # self.keep_alive_task.cancel()
            if event:
                self.events['unchoked'] = event
            self.writer.write(struct.pack('>IB', 1, 2)) # <len=1><id=2>
            await self.writer.drain()
            # self.keep_alive_task = asyncio.get_event_loop().call_later(KEEP_ALIVE_TIMEOUT, asyncio.create_task, self.keep_alive())
        except Exception as e:
            await self.drop()
    
    async def keep_alive(self):
        self.last_active = asyncio.get_event_loop().time()
        # self.keep_alive_task.cancel()
        try:
            self.writer.write(struct.pack('>I', 0)) # <len=0>
            await self.writer.drain()
        except Exception as e:
            self.healthy = False
            return
        # self.keep_alive_task = asyncio.get_event_loop().call_later(KEEP_ALIVE_TIMEOUT, asyncio.create_task, self.keep_alive())
    
    async def drop(self):
        # print('dropping')
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
        if self.keep_alive_task:
            self.keep_alive_task.cancel()
        for i in self.events:
            if i != 'unchoked':
                self.events[i].set_result(b'')
                # TODO: set_exception
        self.events = {}
        self.healthy = False

    def is_busy(self):
        return bool(self.events)



def _decode_str(bencode_str: bytes):
    delim = bencode_str.find(b':')
    size = int(bencode_str[:delim])
    string = bencode_str[delim + 1: delim + size + 1]
    try:
        string = string.decode()
    except UnicodeDecodeError:
        pass
    return string, delim + size + 1

def _decode_int(bencode_str: bytes):
    end = bencode_str.find(b'e')
    return int(bencode_str[1: end]), end+1

def _decode_list(bencode_str: bytes):
    ans = []
    bencode_str = bencode_str[1:]
    processed = 1
    while bencode_str and not bencode_str.startswith(b'e'):
        elem, size = _decode_unknown(bencode_str)
        bencode_str = bencode_str[size:]
        processed += size
        ans.append(elem)
    return ans, processed + 1

def _decode_dict(bencode_str: bytes):
    ans = {}
    bencode_str = bencode_str[1:]
    processed = 1
    while bencode_str and not bencode_str.startswith(b'e'):
        key, keysize = _decode_str(bencode_str)
        bencode_str = bencode_str[keysize:]
        val, valsize = _decode_unknown(bencode_str)
        bencode_str = bencode_str[valsize:]
        processed += keysize + valsize
        ans[key] = val
    return ans, processed + 1

def _decode_unknown(bencode_str: bytes):
    if bencode_str.startswith(b'i'):
        return _decode_int(bencode_str)
    elif bencode_str.startswith(b'l'):
        return _decode_list(bencode_str)
    elif bencode_str.startswith(b'd'):
        return _decode_dict(bencode_str)
    else:
        return _decode_str(bencode_str)

def decode(bencode_str: bytes):
    return _decode_unknown(bencode_str)[0]

def encode(obj):
    if isinstance(obj, bytes):
        return bytes(str(len(obj)), encoding='utf8') + b':' + obj
    elif isinstance(obj, str):
        return bytes(str(len(obj)), encoding='utf8') + b':' + bytes(obj, encoding='utf8')
    elif isinstance(obj, int):
        return b'i' + bytes(str(obj), encoding='utf8') + b'e'
    elif isinstance(obj, list):
        return b'l' +  b''.join(encode(i) for i in obj) + b'e'
    elif isinstance(obj, dict):
        return b'd' + b''.join(encode(i) + encode(obj[i]) for i in obj) + b'e'

if __name__ == '__main__':
    a = b"d8:intervali3600e5:peersld2:ip13:192.168.24.524:porti2001eed2:ip11:192.168.0.34:porti6889eeee"
    b = decode(a)
    print(b)
    c = encode(b)
    print(c)
    print(c == a)



import asyncio
import random

BLOCK_LENGTH = 2**14 # 16KB
UNCHOKE_TIMEOUT = 10
MAX_TRACKERS = 10

class PeerManager:
    def __init__(self, trackers, torrent, max_conn):
        self.trackers = trackers
        self.torrent = torrent
        self.max_peers = max_conn
        self.tracker_semaphore = asyncio.Semaphore(MAX_TRACKERS)
        self.peers = asyncio.Queue()
        self.active_peers = []
        self.is_ready = asyncio.Event()
        self.peers_semaphore = asyncio.Semaphore(max_conn)
    
    async def wait_ready(self):
        # print('waiting for peers')
        await self.check_ready()
        await self.is_ready.wait()
        # print('wait ended')
    
    async def check_ready(self):
        i = 0
        while i < len(self.active_peers):
            if not self.active_peers[i].healthy or self.active_peers[i].am_choking:
                await self.remove_peer(self.active_peers[i])
            else:
                i += 1
        for i in self.active_peers:
            if not i.is_busy():
                self.is_ready.set()
                return
        self.is_ready.clear()
    
    async def remove_peers_not_having_piece(self, pieces):
        i = 0
        while i < len(self.active_peers):
            peer = self.active_peers[i]
            for piece in pieces:
                if peer.pieces[piece]:
                    break
            else:
                await self.remove_peer(peer)
                continue
            i += 1
        await self.check_ready()
    
    async def capture_peers(self):
        for tracker in self.trackers:
            asyncio.create_task(self.add_peers(tracker))
    
    async def add_peers(self, tracker):
        async with self.tracker_semaphore:
            try:
                peers = await tracker.get_peer_list()
                for peer in peers:
                    await self.peers.put(peer)
            except Exception as e:
                pass
                # print('error capturing from tracker', e, type(e))
    
    async def ensure_peers(self):
        await self.capture_peers()
        while True:
            await self.peers_semaphore.acquire() # so we dont get more than "max_conn" peers at the same time
            if self.tracker_semaphore._value == MAX_TRACKERS and self.peers.qsize() == 0:
                await self.capture_peers()
            peer = await self.peers.get()
            asyncio.create_task(self.check_peer(peer))
    
    async def check_peer(self, peer):
        try:
            await peer.handshake()
            if peer.healthy:
                event = asyncio.Event()
                await peer.show_interest(event)
                await asyncio.wait_for(event.wait(), timeout=UNCHOKE_TIMEOUT)
                self.active_peers.append(peer)
                await self.check_ready()
        except Exception as e:
            # print('error handshaking with peer', e, type(e))
            await peer.drop()
            self.peers_semaphore.release()
    
    async def remove_peer(self, peer):
        self.active_peers.remove(peer)
        await peer.drop()
        self.peers_semaphore.release()
        await self.check_ready()
    
    async def peer_having_piece(self, index):
        peers = []
        for peer in self.active_peers:
            if peer.healthy and not peer.am_choking and not peer.is_busy() and peer.pieces[index]:
                peers.append(peer)
        if peers:
            return random.choice(peers)
        return None
    
    async def get_piece_from_peer(self, index, peer, piece_length):
        futures = []
        for begin in range(0, piece_length, BLOCK_LENGTH):
            future = asyncio.Future()
            if begin + BLOCK_LENGTH > piece_length:
                await peer.request_block(index, begin, piece_length - begin, future)
            else:
                await peer.request_block(index, begin, BLOCK_LENGTH, future)
            futures.append(future)
            # print('requested from', begin)
        await self.check_ready()
        return self.reconstruct_piece(futures)
    
    async def reconstruct_piece(self, future_blocks):
        asd = future_blocks
        i = 0
        while asd:
            d, asd = await asyncio.wait(asd, return_when=asyncio.FIRST_COMPLETED)
            i += len(d)
            # self.torrent.progress_bat.update(len(d) * BLOCK_LENGTH)
            # print(f'\033[92mgot {i} part out of {len(future_blocks)}' + '\033[0m')
        blocks = await asyncio.gather(*future_blocks)
        piece = b''
        for block in blocks:
            piece += block
        await self.check_ready()
        return piece



import asyncio
import struct
import random
from urllib.parse import quote_plus, urlparse

from aiohttp import ClientSession

import bencode
from peer import Peer

TRACKER_TIMEOUT = 15

user_agent = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:134.0) Gecko/20100101 Firefox/134.0'}

class Tracker:
    def __init__(self, announce_addr, torrent):
        self.torrent = torrent
        self.announce_addr = announce_addr
        self.peer_list = []
    
    async def get_peer_list(self):
        if self.announce_addr.startswith('http'):
            return await self._get_http_peer_list()
        elif self.announce_addr.startswith('udp'):
            # print('calling udp')
            return await self._get_udp_peer_list()
        else:
            raise NotImplementedError('Unsupported protocol')
    
    async def _get_udp_peer_list(self):
        loop = asyncio.get_running_loop()
        on_con_lost = loop.create_future()
        host, port = urlparse(self.announce_addr).netloc.split(':')
        # print(host, port)
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: _UdpTrackerProtocol(self, on_con_lost),
            remote_addr=(host, port))

        try:
            await asyncio.wait_for(on_con_lost, TRACKER_TIMEOUT)
        finally:
            transport.close()
        # print('reel peers', len(self.peer_list))
        return self.peer_list
        
    async def _get_http_peer_list(self):
        async with ClientSession(headers=user_agent) as session:
            r = await session.get(
                self.announce_addr + f'?info_hash={quote_plus(self.torrent.info_hash)}',
                params={
                        'peer_id': self.torrent.peer_id,
                        'port': 6881,
                        'uploaded': 0,
                        'downloaded': 0,
                        'left': self.torrent.total_length,
                        # 'compact': 1, #TODO
                        'event': 'started',
                        }, 
                timeout=TRACKER_TIMEOUT) #TODO
            answer = bencode.decode(await r.read())
            self.peer_list = answer.get('peers', [])
            if isinstance(self.peer_list, bytes):
                self.peer_list = self.bytes_to_peers(self.peer_list)
            else:
                self.peer_list = [Peer(peer['ip'], peer['port'], self.torrent) for peer in self.peer_list]
            return self.peer_list
    
    def bytes_to_peers(self, data):
        while data:
            *ip, port = struct.unpack('>BBBBH', data[:6])
            self.peer_list.append(Peer('.'.join(str(i) for i in ip), port, self.torrent))
            data = data[6:]

class _UdpTrackerProtocol:
    def __init__(self, tracker: Tracker, on_con_lost):
        self.tracker = tracker
        self.on_con_lost = on_con_lost
        self.state = 'connection'
        self.transport = None

    def connection_made(self, transport):
        self.transport = transport
        self._send_connection_request()

    def datagram_received(self, data, addr):
        if self.state == 'connection':
            if len(data) < 16:
                # print('Invalid data received')
                return
            action, transaction_id, connection_id = struct.unpack('>IIQ', data[:16])
            if action == 0 and transaction_id == self.transaction_id:
                # print('connection response received')
                self._send_announce_request(connection_id)
        elif self.state == 'announce':
            if len(data) < 20:
                # print('Invalid data received')
                return
            action, transaction_id, interval, leechers, seeders = struct.unpack('>IIIII', data[:20])
            if action == 1 and transaction_id == self.transaction_id:
                # print('announce response received')
                # print('interval:', interval)
                # print('leechers:', leechers)
                # print('seeders:', seeders)
                data = data[20:]

                self.tracker.bytes_to_peers(data)

            self.on_con_lost.set_result(True)
        else:
            pass
            # print('Invalid action or transaction_id')

        self.data = data
    
    def eof_received(self):
        pass
        # print("EOF received")

    def error_received(self, exc):
        pass
        # print('Error received:', exc)

    def connection_lost(self, exc):
        pass
        # print("Connection closed")
    
    def _send_connection_request(self):
        # print('sending connection request')
        self.transaction_id = random.randint(0, 0xFFFFFFFF)
        connection_request = struct.pack('>QII',
                                         0x41727101980, # protocol_id, magic constant
                                         0, # action, connect
                                         self.transaction_id)
        self.transport.sendto(connection_request)
    
    def _send_announce_request(self, connection_id):
        # print('sending announce request')
        self.state = 'announce'
        self.transaction_id = random.randint(0, 0xFFFFFFFF)
        info_hash = self.tracker.torrent.info_hash
        announce_request = struct.pack('>QII20s20sQQQiiiiH',
                                        connection_id,
                                        1, # action, announce
                                        self.transaction_id,
                                        info_hash,
                                        self.tracker.torrent.peer_id.encode(),
                                        0, # downloaded
                                        self.tracker.torrent.total_length, # left
                                        0, # uploaded
                                        2, # event, started
                                        0, # ip
                                        90, # key
                                        -1, # num_want
                                        6881, # port
                                        )
        self.transport.sendto(announce_request)

async def main():
    pass

if __name__ == '__main__':
    asyncio.run(main())
    
